#
# $Id:$
#
# This file is part of the Cloud Services Integration Platform (CSIP),
# a Model-as-a-Service framework, API, and application suite.
#
# 2012-2020, OMSLab, Colorado State University.
#
# OMSLab licenses this file to you under the MIT license.
# See the LICENSE file in the project root for more information.
#

from ..utils import utils
from .csip_access import csip_worker
from pyswarms.single.global_best import GlobalBestPSO
from os import path
from threading import Thread
from typing import Dict, List, Set, Tuple
import numpy as np
import copy
import datetime
import queue
import json
import os
from multiprocessing import Queue as MPQueue


def eval_cost(x, iteration, step_param_names, step_objfunc, calib_params, req_queue, files, url, param, conf: Dict, rnd,
              step):
    particles = len(x[:, 0])

    pfail_count = conf.get('particles_fail', 1)  # Number of particles allowed to fail.
    pfail_retry = conf.get('particles_retry', 3)  # retry number of times if more than allowed fail

    while pfail_retry > 0:
        cost = np.ones(particles)
        res_queue = queue.Queue()

        print('   ', end='', flush=True)

        # submit for processing
        # for i_particle, v in enumerate(x[:, 0]):
        for particle in range(particles):
            req_queue.put((rnd, step, iteration, particle, x, step_param_names, calib_params, step_objfunc, res_queue))
            # req_queue.put((i_particle, x[i_particle,:], step_param_names, calib_params, step_objfunc, res_queue))

        # wait for the cost value to come back
        # for i, v in enumerate(x[:, 0]):
        for idx in range(particles):
            (particle, p_cost) = res_queue.get()
            cost[particle] = p_cost
            res_queue.task_done()

        res_queue.join()

        # replace the 'nan' cost values (failed/missing runs) with the mean of the
        # rest of the cost values, hence ignore it

        # print("cost ", cost)
        nan_idx = np.where(np.isnan(cost))
        failed_particles = len(nan_idx[0])

        # leave the loop if fails acceptable
        if failed_particles <= pfail_count:
            break
        
        print("Re-running particles, since ", failed_particles, ' out of ', particles, ' particles failed.')
        pfail_retry -= 1

    if pfail_retry == 0:
        print('Particle evaluation failed ', conf.get('particles_retry', 3), ' times. PSO stopped.')
        return None

    # print("mean ", mean)
    # assign the mean value to all failed runs.
    mean = np.nanmean(cost)
    cost[nan_idx[0]] = mean

    print(flush=True)
    return cost

def global_best(steps: Dict, rounds: Tuple, args: Dict, n_particles: int, iters: int, options: Dict,
                oh_strategy: Dict = None, n_threads: int = 4, rtol: float = 0.001, ftol: float = -np.inf,
                ftol_iter: int = 1, full_trace: List = None, rtol_iter: int = 1,
                conf: Dict = None, metainfo: Dict = None, cost_target: float = -np.inf, result_queue: MPQueue = None) -> Tuple:
    """Performs a stepwise particle swarm optimization PSO using a global best approach.

        Parameters
        ----------
        steps : Dict
            step definitions
        rounds : tuple
            round definition,  (min,max) or max
        args : Dict
            static service args
        n_particles : int
            number of particles
        iters : int
            number of iterations
        options : Dict
            PSO options (see pyswarms)
        oh_strategy : Dict
            PSO Option handling strategy (see pyswarms)
        n_threads : int
            size of thread pool (default: 4)
        rtol : float
            percentage of change of sum(best_cost) between rounds for
            convergence. (Default is 0.001 0.1%)
        ftol : float
            PSO tolerance (default: -np.inf)
        ftol_iter : float
            number of iterations over which the relative error in
            objective_func is acceptable for convergence. (default: 1)
        full_trace : List
            trace of all runs, list of tuples
            first is dictionary of parameter names to parameter values
            second is the cost value (default: None)
        rtol_iter : int
            the number of subsequent rounds with sum(best_cost) < rtol
            (default: 1)
        conf : Dict
            configuration settings (default: {} )
        metainfo : Dict
             additional metainfo for the csip client (default: {} )
        cost_target: float
             the cost target (default: -np.inf)
        Returns
        -------
        Tuple
            optimizer: List, step_trace: Dict

    """

    utils.check_url(args['url'])

    step_file = conf.get('step_trace', None)

    min_rounds = 1
    if type(rounds) == tuple:
        min_rounds = rounds[0]
        max_rounds = rounds[1]
    else:
        max_rounds = rounds

    if min_rounds < 1:
        raise Exception('min rounds >= 1 expected, was "{}"'.format(min_rounds))

    if max_rounds > 20:
        raise Exception('max rounds <= 20 expected, was "{}"'.format(max_rounds))

    if n_threads < 1:
        raise Exception('n_threads >= 1, was "{}"'.format(n_threads))

    if rtol_iter < 1:
        raise Exception('rtol_iter >= 1, was "{}"'.format(rtol_iter))

    if full_trace is not None and not isinstance(full_trace, list):
        raise Exception('full_trace must be of type, was "{}"'.format(type(full_trace)))

    best_cost = np.ones(len(steps)) * np.inf
    optimizer = np.empty(len(steps), dtype=object)

    # trace of steps info
    step_trace = {}

    step_trace['dir'] = os.getcwd()
    step_trace['start'] = str(datetime.datetime.now())
    step_trace['min_rounds'] = min_rounds
    step_trace['max_rounds'] = max_rounds
    step_trace['iters'] = iters
    
    # BUG If ftol is -inf set it to a string
    ftol_value = ftol
    if ftol == -np.inf:
        ftol_value = '-inf'
    elif ftol == np.inf:
        ftol_value = 'inf'
    
    step_trace['ftol'] = ftol_value
    step_trace['ftol_iter'] = ftol_iter
    step_trace['rtol'] = rtol
    step_trace['rtol_iter'] = rtol_iter
    step_trace['n_threads'] = n_threads
    step_trace['n_particles'] = n_particles
    step_trace['n_steps'] = len(steps)
    step_trace['steps'] = copy.deepcopy(steps)
    step_trace['args'] = str(args) #BUG MUST BE REMOVED

    if step_file is not None:
        with open(step_file, "w") as fo:
            json.dump(step_trace, fo)

    print("Wrote step trace")

    # best round cost
    best_round_cost = np.inf

    # request queue for worker
    req_queue = queue.Queue()
    
    print("Created queue")

    conf = conf or {}
    done = False
    thread_pool = []
    for thread_no in range(n_threads):
        worker = Thread(target=csip_worker, args=(req_queue, thread_no, lambda: done,
                                                  full_trace, args['url'], args.get('files', None), args['param'],
                                                  conf, metainfo)
                        )
        thread_pool.append(worker)
        worker.start()

    print("Started worker threads")

    r_below = 0
    early_exit = False
    start_time = datetime.datetime.now()
    for r in range(max_rounds):
        no_improvement = np.full(len(steps), True)
        for s, step in enumerate(steps):

            # check if forced exit.
            if path.exists("stop"):
                print('\n>>>>> stop file found, exit now.')
                early_exit = True
                break

            param_names, bounds, objfunc = utils.get_step_info(steps, s)
            # maybe clone args?
            args['step_param_names'] = param_names
            args['step_objfunc'] = objfunc
            # get calibrated parameter from all other steps
            args['calib_params'] = utils.get_calibrated_params(steps, s)

            args['req_queue'] = req_queue
            args['conf'] = conf

            print("Calling global best..")

            # create optimizer in the first round.
            if optimizer[s] is None:
                optimizer[s] = GlobalBestPSO(step.get('n_particles', n_particles),
                                             len(param_names),
                                             oh_strategy=step.get('oh_strategy', oh_strategy),
                                             options=step.get('options', options),
                                             bounds=bounds,
                                             ftol=step.get('ftol', ftol),
                                             ftol_iter=step.get('ftol_iter', ftol_iter),
                                             cost_target=step.get('cost_target', cost_target))
                
            print('\n>>>>> R{}/S{}  particle params: {}  calibrated params: {}\n'.format(r + 1, s + 1, param_names, args['calib_params']))

            #if result_queue is not None:
            #    result_queue.put('\n>>>>> R{}/S{}  particle params: {}  calibrated params: {}\n'.format(r + 1, s + 1, param_names, args['calib_params']))
            
            print("Filled request queue...")
            
            args['rnd'] = r + 1
            args['step'] = s + 1

            print("Evaluating cost...")

            # perform optimization
            cost, pos = optimizer[s].optimize(eval_cost, iters=step.get('iters', iters), **args)
            if cost is None:
                early_exit = True
                break

            print("Finished evaluation...")

            # capture the best cost
            # if cost < best_cost[s] and np.abs(cost - best_cost[s]) > rtol:
            if cost < best_cost[s]:
                best_cost[s] = cost
                no_improvement[s] = False
                utils.annotate_step(best_cost[s], pos, steps, s)

            print('\n     Step summary, best particle values: {} '.format(pos))
            
            if result_queue is not None:
                result_queue.put('\n     Step summary, best particle values: {} '.format(pos))

            key = "r{}s{}".format(r + 1, s + 1)
            step_trace[key] = {}
            step_trace[key]['time'] = str(datetime.datetime.now())
            
            best_costs_list = best_cost.tolist()
            # If the cost is inf, set it to a string
            for i, c in enumerate(best_costs_list):
                if c == np.inf:
                    best_costs_list[i] = 'inf'
                elif c == -np.inf:
                    best_costs_list[i] = '-inf'
            
            step_trace[key]['best_costs'] = best_costs_list # BUG
            step_trace[key]['steps'] = copy.deepcopy(steps)

            if step_file is not None:
                with open(step_file, "w") as fo:
                    json.dump(step_trace, fo)

            # print(json.dumps(steps, sort_keys=False, indent=2))

        if early_exit:
            step_trace['exit'] = '1'
            break

        round_cost = np.sum(best_cost)

        # if no improvement in all steps, break out of rounds prematurely
        # but start checking only after min_rounds
        # if (r + 1 >= min_rounds) and all(no_improvement):
        rel_round_tol = 1 - round_cost / best_round_cost

        print('\n  Round summary - round_cost:{}, step_costs: {}, step improvement:{}'
              .format(round_cost, best_cost, np.invert(no_improvement)))
        print('\n  Progress -  best_round_cost:{}, rel_round_tol:{}, rtol:{}'
              .format(best_round_cost, rel_round_tol, rtol))

        if result_queue is not None:
            result_queue.put('\n  Round summary - round_cost:{}, step_costs: {}, step improvement:{}'
              .format(round_cost, best_cost, np.invert(no_improvement)))
            
        if result_queue is not None:
            result_queue.put('\n  Progress -  best_round_cost:{}, rel_round_tol:{}, rtol:{}'
              .format(best_round_cost, rel_round_tol, rtol))

        key = "r{}".format(r + 1)
        step_trace[key] = {}
        step_trace[key]['time'] = str(datetime.datetime.now())
        step_trace[key]['round_cost'] = round_cost
        
        best_costs_list = best_cost.tolist() #BUG
        # If the cost is inf, set it to a string
        for i, c in enumerate(best_costs_list):
            if c == np.inf:
                best_costs_list[i] = 'inf'
            elif c == -np.inf:
                best_costs_list[i] = '-inf'
        
        step_trace[key]['best_costs'] = best_costs_list
        step_trace[key]['improvements'] = no_improvement.tolist()
        if step_file is not None:
            with open(step_file, "w") as fo:
                json.dump(step_trace, fo)

        if (r + 1 >= min_rounds) and 0 <= rel_round_tol < rtol:
            r_below += 1
            if r_below >= rtol_iter:
                break
        else:
            # reset
            r_below = 0

        if round_cost < best_round_cost:
            best_round_cost = round_cost

    end_time = datetime.datetime.now()
    elapsed = str(end_time - start_time)

    print('Done in {} after {} out of {} rounds'.format(elapsed, r + 1, max_rounds))
    
    if result_queue is not None:
        result_queue.put('Done in {} after {} out of {} rounds'.format(elapsed, r + 1, max_rounds))

    done = True
    for worker in thread_pool:
        worker.join()

    step_trace['rounds'] = r + 1
    step_trace['end'] = str(datetime.datetime.now())
    step_trace['time'] = elapsed

    if step_file is not None:
        with open(step_file, "w") as fo:
            json.dump(step_trace, fo)

    if result_queue is not None:
        result_queue.put("Step Trace")
        result_queue.put(step_trace)

    return optimizer, step_trace



"""
CTkToolTip Widget
version: 0.4
"""

import time
import sys
import customtkinter
from tkinter import Toplevel, Frame

class CTkToolTip(Toplevel):
    """
    Creates a ToolTip (pop-up) widget for customtkinter.
    """

    def __init__(
        self,
        widget: any = None,
        message: str = None,
        delay: float = 0.2,
        follow: bool = True,
        x_offset: int = +20,
        y_offset: int = +10,
        bg_color: str = None,
        corner_radius: int = 10,
        border_width: int = 0,
        border_color: str = None,
        alpha: float = 0.8,
        padding: tuple = (10,2),
        **message_kwargs):
        
        super().__init__()

        self.widget = widget
        
        self.withdraw()
        # Disable ToolTip's title bar
        self.overrideredirect(True)
                
        if sys.platform.startswith("win"):
            self.transparent_color = self.widget._apply_appearance_mode(customtkinter.ThemeManager.theme["CTkToplevel"]["fg_color"])
            self.attributes("-transparentcolor", self.transparent_color)
            self.transient()
        elif sys.platform.startswith("darwin"):
            self.transparent_color = 'systemTransparent'
            self.attributes("-transparent", True)
            self.transient(self.master)
        else:
            self.transparent_color = '#000001'
            corner_radius = 0
            self.transient()
            
        self.resizable(width=True, height=True)
        self.transient()
        
        # Make the background transparent
        self.config(background=self.transparent_color)
        
        # StringVar instance for msg string
        self.messageVar = customtkinter.StringVar()
        self.message = message
        self.messageVar.set(self.message)
      
        self.delay = delay
        self.follow = follow
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.corner_radius = corner_radius
        self.alpha = alpha
        self.border_width = border_width
        self.padding = padding
        self.bg_color = bg_color
        self.border_color = border_color
        self.disable = False
        
        # visibility status of the ToolTip inside|outside|visible
        self.status = "outside"
        self.last_moved = 0
        self.attributes('-alpha', self.alpha)
        
        # Add the message widget inside the tooltip
        self.transparent_frame = Frame(self, bg=self.transparent_color)
        self.transparent_frame.pack(padx=0, pady=0, fill="both", expand=True)
        
        self.frame = customtkinter.CTkFrame(self.transparent_frame, bg_color=self.transparent_color, corner_radius=self.corner_radius,
                                            border_width=self.border_width, fg_color=self.bg_color, border_color=self.border_color)
        self.frame.pack(padx=0, pady=0, fill="both", expand=True)
        
        self.message_label = customtkinter.CTkLabel(self.frame, textvariable=self.messageVar, **message_kwargs)
        self.message_label.pack(fill="both", padx=self.padding[0]+self.border_width,
                                pady=self.padding[1]+self.border_width, expand=True)

        if self.widget != None:
            if self.widget.winfo_name()!="tk":
                if self.frame.cget("fg_color")==self.widget.cget("bg_color"):
                    if not bg_color:             
                        self._top_fg_color = self.frame._apply_appearance_mode(customtkinter.ThemeManager.theme["CTkFrame"]["top_fg_color"])
                        self.frame.configure(fg_color=self._top_fg_color)

            # Add bindings to the widget without overriding the existing ones
            self.widget.bind("<Enter>", self.on_enter, add="+")
            self.widget.bind("<Leave>", self.on_leave, add="+")
            self.widget.bind("<Motion>", self.on_enter, add="+")
            self.widget.bind("<B1-Motion>", self.on_enter, add="+")
            self.widget.bind("<Destroy>", lambda _: self.hide(), add="+")

    def show(self) -> None:
        """
        Enable the widget.
        """
        self.disable = False
        
    def on_enter(self, event) -> None:
        """
        Processes motion within the widget including entering and moving.
        """

        if self.disable: return
        self.last_moved = time.time()

        # Set the status as inside for the very first time
        if self.status == "outside":
            self.status = "inside"

        # If the follow flag is not set, motion within the widget will make the ToolTip dissapear
        if not self.follow:
            self.status = "inside"
            self.withdraw()

        # Offsets the ToolTip using the coordinates od an event as an origin
        self.geometry(f"+{event.x_root + self.x_offset}+{event.y_root + self.y_offset}")

        # Time is in integer: milliseconds
        self.after(int(self.delay * 1000), self._show)

    def on_leave(self, event=None) -> None:
        """
        Hides the ToolTip temporarily.
        """

        if self.disable: return
        self.status = "outside"
        self.withdraw()

    def _show(self) -> None:
        """
        Displays the ToolTip.
        """

        if not self.widget.winfo_exists():
            self.hide()
            self.destroy()
            
        if self.status == "inside" and time.time() - self.last_moved >= self.delay:
            self.status = "visible"
            self.deiconify()
        
    def hide(self) -> None:
        """
        Disable the widget from appearing.
        """
        if not self.winfo_exists():
            return
        self.withdraw()
        self.disable = True

    def is_disabled(self) -> None:
        """
        Return the window state
        """
        return self.disable
    
    def get(self) -> None:
        """
        Returns the text on the tooltip.
        """  
        return self.messageVar.get()
    
    def configure(self, message: str = None, delay: float = None, bg_color: str = None, **kwargs):
        """
        Set new message or configure the label parameters.
        """
        if delay: self.delay = delay
        if bg_color: self.frame.configure(fg_color=bg_color)
        
        self.messageVar.set(message)
        self.message_label.configure(**kwargs)



from customtkinter import CTkScrollableFrame
from customtkinter import CTkFrame
from customtkinter import CTkLabel
from customtkinter import CTkButton
from customtkinter import CTkEntry
from customtkinter import CTkTextbox
from customtkinter import CTkImage
from customtkinter import CTkOptionMenu
from PIL import Image
import os

from . import BoundsList
from . import FunctionsList
from . import ListEditor


class StepView(CTkScrollableFrame):
    def __init__(self, *args,
                 option_manager: None,
                 home_page: None,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.option_manager = option_manager
        self.analysisFrame = None
        self.containerFrame = None

        self.render()

    def refresh(self):
        self.clear()
        self.render()

    def clear(self):
        if self.containerFrame != None:
            self.containerFrame.destroy()

        if self.analysisFrame != None:
            self.analysisFrame.destroy()

    def create_new_step(self):

        self.clear()
        example_step = [{    # step 1
            'param': [
                {
                    'name': 'soilOutLPS',
                    'bounds': (0.0, 2.0)
                },
                {
                    'name': 'lagInterflow',
                    'bounds': (10.0, 80.0)
                }
            ],
            'objfunc': [
                {
                    'name': 'ns',
                    'of': 'ns',
                    'data': ('obs_data02_14.csv/obs/orun[1]',
                             'output/csip_run/out/Outlet.csv/output/catchmentSimRunoff')
                }
            ],
            'open': True
        }]
        self.option_manager.add_steps(example_step)

        self.render()

    def render(self):

        row = 0
        index = 0

        self.containerFrame = CTkFrame(self)
        self.containerFrame.grid(row=0, column=0, padx=(
            10, 10), pady=(10, 10), sticky="nsew")
        self.containerFrame.grid_columnconfigure((0, 1), weight=1)

        self.steps = self.option_manager.get_steps()
        self.mode = self.option_manager.get_arguments()['mode'].get()

        if (self.mode == "Sensitivity Analysis"):
            info = self.option_manager.get_project_data()
            folder = os.path.join(info['path'], info['name'], "results")
            
            # File all CSV files in folder and put them into list with strings as path
            files = []
            if os.path.exists(folder):
                files = [f for f in os.listdir(folder) if f.endswith('.csv')]

            if len(files) == 0:
                files = ["      No CSV files found! Run sampling first!       "]

            header_padding_x = (5, 5)
            header_padding_y = (10, 10)

            self.file_selector_frame = CTkFrame(self.containerFrame, corner_radius=0)
            self.file_selector_frame.grid(row=0, column=0, sticky="nsew", columnspan=2)
            self.logo_label = CTkLabel(self.file_selector_frame, text="Select File:")
            self.logo_label.grid(row=0, column=0, padx=(20, 10), pady=header_padding_y)
            
            self.file_selector = CTkOptionMenu(self.file_selector_frame, values=files, width=50)
            self.file_selector.grid(row=0, column=1, padx=(5, 80), pady=header_padding_y)
            self.file_selector.set(files[0])

            row += 1


        for step in self.steps:

            up_image = CTkImage(Image.open(os.path.join("./images", "up.png")), size=(20, 20))
            down_image = CTkImage(Image.open(os.path.join("./images", "down.png")), size=(20, 20))
            trash_image = CTkImage(Image.open(os.path.join("./images", "trash.png")), size=(20, 20))
            expand_image = CTkImage(Image.open(os.path.join("./images", "expand.png")), size=(20, 20))
            collapse_image = CTkImage(Image.open(os.path.join("./images", "collapse.png")), size=(20, 20))


            expand_func = lambda index=index: (self.clear(), self.option_manager.toggle_step_open(index), self.render())
            up_func = lambda index=index: (self.clear(), self.option_manager.move_step_up(index), self.render())
            down_func = lambda index=index: (self.clear(), self.option_manager.move_step_down(index), self.render())
            remove_func = lambda index=index: (self.clear(), self.option_manager.remove_step(index), self.render())
            
            if (self.mode == "Optimization: MG-PSO"):
                button_container = CTkFrame(self.containerFrame, width=200)
                button_container.grid(row=row, column=1, sticky="nse", padx=(10, 10), pady=(10, 10))
                button_container.grid_rowconfigure(0, weight=1)
                button_container.grid_columnconfigure(0, weight=1)
                
                CTkEntry(self.containerFrame, textvariable=step['name'], width=500).grid(row=row, column=0, padx=(20, 20), pady=(20, 20), sticky="nsw")
            
                CTkButton(button_container, width=30, text=None, image=expand_image if not step['open'] else collapse_image, command=expand_func).grid(row=0, column=0, padx=(10, 10), pady=(10, 10), sticky="nsew")
                CTkButton(button_container, width=30, text=None, image=up_image, state="disabled" if index==0 else "normal", fg_color="gray" if index==0 else None, command=up_func).grid(row=0, column=1, padx=(10, 10), pady=(10, 10), sticky="nsew")
                CTkButton(button_container, width=30, text=None, image=down_image, state="disabled" if index==(len(self.steps)-1) else "normal", fg_color="gray" if index==(len(self.steps)-1) else None, command=down_func).grid(row=0, column=2, padx=(10, 10), pady=(10, 10), sticky="nsew")
                CTkButton(button_container, width=30, text=None, image=trash_image, command=remove_func).grid(row=0, column=3, padx=(10, 10), pady=(10, 10), sticky="nsew")

                row += 1

            if step['open'] or (self.mode == "Sampling: Halton" or self.mode == "Sampling: Random" or self.mode == "Sensitivity Analysis"):
                bounds = BoundsList.BoundsList(
                    self.containerFrame, option_manager=self.option_manager, step_index=index)
                bounds.grid(row=row, column=0, padx=(10, 10),
                            pady=(10, 10), sticky="nsew")
                bounds.grid_columnconfigure(0, weight=1)
                bounds.grid_rowconfigure(0, weight=1)
                
                funcs = FunctionsList.FunctionsList(
                    self.containerFrame, option_manager=self.option_manager, step_index=index)
                funcs.grid(row=row, column=1, padx=(10, 10),
                            pady=(10, 10), sticky="nsew")
                funcs.grid_columnconfigure(0, weight=1)
                funcs.grid_rowconfigure(0, weight=1)
                
            row += 1
            index += 1
            
            if (self.mode != "Optimization: MG-PSO"):
                break

        # Create an "Add step button that is centered
        if (self.mode == "Optimization: MG-PSO" or len(self.steps) == 0):
            CTkButton(self.containerFrame, text="Add Group", command=self.create_new_step).grid(
                row=row, columnspan=2, column=0, padx=(10, 10), pady=(10, 10), sticky="ew")
            
        """
        if (self.mode == "Sampling: Halton" or self.mode == "Sampling: Random" or self.mode == "Sensitivity Analysis"):
            self.analysisFrame = CTkFrame(self)
            self.analysisFrame.grid(row=2, column=0, columnspan=2, padx=(10, 10), pady=(10, 10), sticky="nsew")
            self.analysisFrame.grid_columnconfigure((0, 1, 2), weight=1)

            columns = ["n1", "n2", "n3", "n4", "n5", "n6", "n7", "n8", "n9", "n10"]

            self.params_editor = ListEditor.ListEditor(self.analysisFrame, option_manager=self.option_manager, 
                                                       title="Parameters",
                                                       columns=columns,
                                                       parameter_name="sensitivity_parameters",
                                                       parameter_remove_func=self.option_manager.remove_sensitivity_parameter,
                                                       parameter_add_func=self.option_manager.add_sensitivity_param)
            self.params_editor.grid(row=0, column=0, padx=(10, 10), pady=(10, 0), sticky="nsew")
            self.params_editor.grid_columnconfigure(0, weight=1)
            self.params_editor.grid_rowconfigure(0, weight=1)


            self.positive_editor = ListEditor.ListEditor(self.analysisFrame, option_manager=self.option_manager, 
                                                       title="Positive Best Metrics",
                                                       columns=columns,
                                                       parameter_name="sensitivity_positiveBestMetrics",
                                                       parameter_remove_func=self.option_manager.remove_sensitivity_positiveBestMetrics,
                                                       parameter_add_func=self.option_manager.add_sensitivity_positiveBestMetrics)
            self.positive_editor.grid(row=0, column=1, padx=(10, 10), pady=(10, 0), sticky="nsew")
            self.positive_editor.grid_columnconfigure(0, weight=1)
            self.positive_editor.grid_rowconfigure(0, weight=1)

            self.negative_editor = ListEditor.ListEditor(self.analysisFrame, option_manager=self.option_manager, 
                                                       title = "Zero Best Metrics",
                                                       columns=columns,
                                                       parameter_name="sensitivity_negativeBestMetrics",
                                                       parameter_remove_func=self.option_manager.remove_sensitivity_negativeBestMetrics,
                                                       parameter_add_func=self.option_manager.add_sensitivity_negativeBestMetrics)
            self.negative_editor.grid(row=0, column=2, padx=(10, 10), pady=(10, 0), sticky="nsew")
            self.negative_editor.grid_columnconfigure(0, weight=1)
            self.negative_editor.grid_rowconfigure(0, weight=1)
        """



# -*- coding: utf-8 -*-

__version__ = '1.2.8'
__built__ = '1.2.8 955b3df883a5b6e63e360bf498da9f38cc0953a8 2022-01-27 10:14 -0700 od'
__author__ = 'Olaf David'
__email__ = 'odavid@colostate.edu'

from .pso import pso
from .sampling import sampling
__all__ = ["pso", "sampling"]

